/**
 * @file modelconverter.hpp
 * @date 2016-06-19
 */

// Copyright (C) 2016 Usagi
// Licensed under the MIT License.

// for Siv3D March 2016

#pragma once

#include "mqo.hpp"
#include "pmd.hpp"

/// <summary>ZW]VCX^XԂ</summary>
Vec3 rev_z(const Vec3& v) {
	Vec3 ret;
	ret.x = v.x;
	ret.y = v.y;
	ret.z = -v.z;
	return ret;
}

/// <summary>s3d̍ގɒǉ</summary>
struct Material2 : s3d::Material {
	/// <summary>1`10-1</summary>
	int toonNumber = 1;
	/// <summary>01</summary>
	int edge = 0;
};

/// <summary>ItZbgƌBގƁB</summary>
struct VertexGroup {
	/// <summary>MeshData̖ʒ_CfbNX̃ItZbg</summary>
	int offset = 0;
	/// <summary>MeshData̖ʒ_CfbNX̌</summary>
	int num() const {
		return indices.size();
	};

	/// <summary>ʒ_</summary>
	Array<int> indices;
};

/// <summary>1[V</summary>
struct MotionNode {
	/// <summary>O</summary>
	String name;

	/// <summary>eCfbNX
	/// -1̏ꍇ̓fm[he</summary>
	int parentIndex = -1;

	/// <summary>Sivł̕sړ</summary>
	Vec3 translation = { 0, 0, 0 };
	/// <summary>Sivł̃[J]
	/// rotationvZ</summary>
	Quaternion rot;

	/// <summary>ʒuinO[o</summary>
	Vec3 globalMov = { 0, 0, 0 };
	/// <summary>]inO[o</summary>
	Quaternion globalRot = Quaternion::Identity();
	/// <summary>XP[inO[o(Ή)</summary>
	Vec3 globalRate = { 1, 1, 1 };

	/// <summary>[VƂĔf镪</summary>
	Vec3 motionMov = { 0, 0, 0 };
	/// <summary>[VƂĔf镪</summary>
	Quaternion motionRot = Quaternion::Identity();
	/// <summary>[VƂĔf镪(Ή)</summary>
	Vec3 motionScale = { 1, 1, 1 };

	/// <summary>ƃ[VfZoړ</summary>
	Vec3 curMov = { 0, 0, 0 };
	/// <summary>ƃ[VfZo]B
	/// ɉ]B</summary>
	Quaternion curRot = Quaternion::Identity();
	/// <summary>ƃ[VfZogk</summary>
	Vec3 curScale = { 1, 1, 1 };

	/// <summary>ގƓ</summary>
	Array<VertexGroup> groups;

	/// <summary>vZ̍ŏIs</summary>
	Mat4x4 GetCurMat() const {
		Quaternion revGlobalRot = Quaternion(-globalRot.component.m128_f32[0],
			-globalRot.component.m128_f32[1],
			-globalRot.component.m128_f32[2],
			globalRot.component.m128_f32[3]);
		/*
		Quaternion revGlobalRot = Quaternion(-globalRot.component.fv.x,
			-globalRot.component.fv.y,
			-globalRot.component.fv.z,
			globalRot.component.fv.w);*/
		Mat4x4 ret = Mat4x4::Translate(-globalMov).rotated(revGlobalRot).rotated(curRot).translated(curMov);
		return ret;
	}
};

/// <summary>.mqo1ObjectɑΉ</summary>
struct ObjectNode : public MotionNode {
	/// <summary>.mqołdepth
	/// -1̓fS</summary>
	int depth = -1;
	/// <summary>XP[͔Ή</summary>
	Vec3 scale = {1,1,1};

	/// <summary>SivłRoll,Pitch,Yaw</summary>
	Vec3 rotation = {0,0,0};


	/// <summary>\
	/// 0̏ꍇ͎q</summary>
	int childIndex = 0;

	/// <summary>_Ɩʒ_CfbNX
	/// ̖ʒ_CfbNX͍ގ</summary>
	MeshData meshData;
};

/// <summary>1̃CX^X</summary>
struct MotionInstance {
	/// <summary>S</summary>
	MotionNode modelNode;
	/// <summary>IuWFNg</summary>
	Array<MotionNode> nodes;

	/// <summary>ǉCfbNXB
	/// 1ft@C̑ΉCfbNXi[B</summary>
	Optional<int> optIdx;

	/// <summary>Object̏ԂvZB
	/// scale͖ 1 1 1 Ƃ݂Ȃ</summary>
	void ComputeCurCoordinates() {
		int num = nodes.size();
		for (int i = 0; i < num; ++i) {
			MotionNode& node = nodes[i];
			// ύXȂ悤ɃRs[
			MotionNode n = this->modelNode;
			int parent = node.parentIndex;
			if (parent >= 0 && parent < num) {
				n = nodes[parent];
			}

			node.curMov = n.curMov
				+ n.curRot * (node.translation + node.motionMov);
			node.curRot = n.curRot * node.rot * node.motionRot;

			//Print(L"ʒu, ", node.curMov);
			//Println(L", ], ", node.curRot);
		}
	}
};

/// <summary>.mqoƃbVf[^Ԃ̕ϊ
/// Siv3D\sB</summary>
class ModelConverter {
public:
	/// <summary>ގ</summary>
	Array<Material2> mats;
	/// <summary>eNX`</summary>
	Array<DynamicTexture> texs;

	/// <summary>bV</summary>
	Array<DynamicMesh> meshes;

	/// <summary>IuWFNgƓ</summary>
	Array<ObjectNode> nodes;

	/// <summary>ŏi</summary>
	ObjectNode modelNode;
	/// <summary>t@C̃pX</summary>
	String path;
	/// <summary>gpBWATCY̕ϊ</summary>
	//double rate = 1.0;

	/// <summary>zNA</summary>
	void clear() {
		this->mats.clear();
		this->texs.clear();
		this->meshes.clear();
		this->nodes.clear();
		this->modelNode = ObjectNode();
		this->path.clear();
	}

	/// <summary>bVf[^𐶐</summary>
	/// <param name="f">Kp</param>
	/// <param name="rate">Wl̔{w肷</param>
	/// <param name="option">ʍގ̈A</param>
	bool ApplyMqo(const MqoFile& f,
			double rate,
			const Array<String>& options) {
		bool ret = true;
		this->clear();

		bool addFaceForDouble = false; // ʍގ̏ꍇɖʒǉ
		int cutType = false;
		for (const auto& option: options) {
			if (option == L"addFaceForDouble") {
				addFaceForDouble = true;
			}
			if (option.startsWith(L"cut")) {
				cutType = s3d::Parse<int>(option.substr(3));
			}
		}

		this->path = f.path;

		// ގ̂߂̃eNX`
		String dir = s3d::FileSystem::ParentPath(f.path);

		for (const auto& m : f.materials) {
			Material2 mat;
			mat.diffuse = m.col * m.dif;
			mat.specular = m.spc_col * m.spc;
			mat.shininess = m.power;
			mat.emission = m.col * m.emi;
			mat.ambient = m.amb_col * m.amb;
			//mat.illum = 1;
			mat.diffuseTextureName = m.tex.value_or(L"");
			mat.ambientTextureName = L"";
			mat.normalTextureName = m.bump.value_or(L"");
			mat.name = m.name;
			if (m.ex2) {
				mat.edge = m.ex2.value().edge;
				mat.toonNumber = m.ex2.value().toon;
			}
			else {
				mat.edge = 0;
				mat.toonNumber = -1;
			}
			this->mats.push_back(mat);

			// ʏeNX`
			bool flag = false;
			if (m.tex) {
				auto name = m.tex.value();
				auto fullPath = name;
				if (fullPath.indexOf(L":") != 1) {
					fullPath = dir + L"/" + name;
					if (FileSystem::Exists(fullPath) == false) {
						fullPath = dir + L"/Texture/" + name;
						if (FileSystem::Exists(fullPath) == false) {
							fullPath = FileSystem::ParentPath(dir) + L"Texture/" + name;
						}
					}
				}

				Image img(fullPath);
				if (img) {
					DynamicTexture dtex(img);
					if (dtex) {
						this->texs.push_back(dtex);
						flag = true;
					}
				}
			}
			if (flag == false) {
				DynamicTexture dtex(32,32, Palette::White); // ftHg
				this->texs.push_back(dtex);
			}


			// ʉ
			flag = false;
			if (m.bump) {
				auto name = m.bump.value();
				auto fullPath = name;
				if (fullPath.indexOf(L":") != 1) {
					fullPath = dir + L"/" + name;
					if (FileSystem::Exists(fullPath) == false) {
						fullPath = dir + L"/Texture/" + name;
						if (FileSystem::Exists(fullPath) == false) {
							fullPath = FileSystem::ParentPath(dir) + L"Texture/" + name;
						}
					}
				}

				Image img(fullPath);
				if (img) {
					DynamicTexture dtex(img);
					if (dtex) {
						//this->texs.push_back(dtex);
						flag = true;
					}
				}
			}
			if (flag == false) {
				DynamicTexture dtex(32, 32, Palette::White); // ftHg
				//this->texs.push_back(dtex);
			}

		}

		for (int i = 0; i < f.objects.size(); ++i) {
			MqoObject obj = f.objects[i];
			ObjectNode node;
			node.name = obj.name;
			node.depth = obj.depth;
			node.translation = rev_z(obj.translation) * rate;
			node.rotation = { Radians(obj.rotation.z),
				-Radians(obj.rotation.y),
				-Radians(obj.rotation.x) };
			node.scale = obj.scale;

			int depth = obj.depth;
			int index = -1;
			// ԋ߂1T
			if (depth > 0) {
				for (int j = 0; j < i; ++j) {
					if (f.objects[j].depth == depth - 1) {
						index = j;
					}
				}
			}
			node.parentIndex = index;
			node.childIndex = 0;
			if (index >= 0 && index < this->nodes.size()) {
				if (this->nodes[index].childIndex <= 0) {
					this->nodes[index].childIndex = i;
				}
			}

			node.rot = Quaternion::RollPitchYaw(
				node.rotation.x,
				node.rotation.y,
				node.rotation.z);

			// ގ
			int matNum = f.materials.size();
			for (int j = 0; j < matNum; ++j) {
				VertexGroup gr;
				gr.offset = 0;
				gr.indices.clear();
				node.groups.push_back(gr);
			}

			for (auto& face : obj.faces) {
				this->cutByScore(obj, face, cutType);
			}

			Vec2 duv(-0.5, -0.5);
			// 򂷂
			if (obj.shading == 0) { // oŎOp`S@

				for (const auto& face : obj.faces) {
					int matIndex = face.materialIndex;
					if (matIndex < 0 || matIndex >= matNum) {
						continue;
					}
					VertexGroup& gr = node.groups[matIndex];

					bool dbls = f.materials[matIndex].dbls;
					bool addDoubleFace = false;
					if (dbls && addFaceForDouble) {
						addDoubleFace = true;
					}

					int uvNum = face.uvs.size();

					for (const auto& tri: face.tris) {
						int baseIndex = node.meshData.vertices.size();
						gr.indices.push_back(baseIndex + 0);
						gr.indices.push_back(baseIndex + 1);
						gr.indices.push_back(baseIndex + 2);

						if (addDoubleFace) {
							gr.indices.push_back(baseIndex + 2);
							gr.indices.push_back(baseIndex + 1);
							gr.indices.push_back(baseIndex + 0);
						}

						for (int k = 0; k < 3; ++k) {
							MeshVertex mv;
							// in Face
							int fi = tri[k];
							// in Object
							int v = face.indices[fi];

							mv.position = rev_z(obj.vertices[v]) * rate;
							mv.texcoord = (fi < uvNum) ? face.uvs[fi] : duv;
							node.meshData.vertices.push_back(mv);
						}
					}

				}

			}
			else {
				// 炩

				// ɖ@vZ
				for (const auto& v : obj.vertices) {
					MeshVertex mv;
					mv.position = rev_z(v) * rate;
					mv.texcoord = duv;
					node.meshData.vertices.push_back(mv);
				}
				for (const auto& face : obj.faces) {
					for (const auto& tri: face.tris) {
						node.meshData.indices.push_back(face.indices[tri[0]]);
						node.meshData.indices.push_back(face.indices[tri[1]]);
						node.meshData.indices.push_back(face.indices[tri[2]]);
					}
				}
				node.meshData.computeNormals();

				// ʂ
				for (const auto& face : obj.faces) {
					int matIndex = face.materialIndex;
					if (matIndex < 0 || matIndex >= matNum) {
						continue;
					}
					VertexGroup& gr = node.groups[matIndex];

					bool dbls = f.materials[matIndex].dbls;
					bool addDoubleFace = false;
					if (dbls && addFaceForDouble) {
						addDoubleFace = true;
					}

					int n = face.indices.size();
					int uvNum = face.uvs.size();
					int v0, v1, v2;
					if (n >= 3) {
						// w蒸_ǉ@
						Array<int> adds;
						for (int k = 0; k < n; ++k) {
							v0 = face.indices[k];

							int realIndex = v0;
							if (k < uvNum) {
								auto uv = node.meshData.vertices[v0].texcoord;
								if (uv.x < 0) { // 
									node.meshData.vertices[v0].texcoord = face.uvs[k];
								}
								else if (uv != face.uvs[k]) {
									MeshVertex add(node.meshData.vertices[v0]);
									add.texcoord = face.uvs[k];
									realIndex = node.meshData.vertices.size(); // O
									node.meshData.vertices.push_back(add);
								}
							}
							adds.push_back(realIndex);
						}
						for (const auto& tri: face.tris) {
							v0 = adds[tri[0]];
							v1 = adds[tri[1]];
							v2 = adds[tri[2]];
							gr.indices.push_back(v0);
							gr.indices.push_back(v1);
							gr.indices.push_back(v2);

							if (addDoubleFace) {
								gr.indices.push_back(v2);
								gr.indices.push_back(v1);
								gr.indices.push_back(v0);
							}

						}
					}

				}

			} // ʂ or 炩


			node.meshData.indices.clear();
			for (int i = 0; i < matNum; ++i) {
				node.groups[i].offset = 0;
				if (i >= 1) {
					node.groups[i].offset = node.groups[i - 1].offset
						+ node.groups[i - 1].num();
				}

				for (auto idx : node.groups[i].indices) {
					node.meshData.indices.push_back(idx);
				}
			}
			if (obj.shading == 0) {
				node.meshData.computeNormals();
			}

			this->nodes.push_back(node);

			DynamicMesh dm(node.meshData);
			this->meshes.push_back(dm);
			
			if (dm.isEmpty()) {
				//Println(L"̃_Ci~bNbV, ",i);
			}

		} // IuWFNg

		this->computeBaseCoordinates();

		return ret;
	}


	/// <summary>face̕߂</summary>
	/// <param name="obj">ʂ܂ރIuWFNg</param>
	/// <param name="face">Ώۂ1</param>
	/// <param name="cutType">B0: PA1: IA2: +ˉeI</param>
	int cutByScore(const MqoObject& obj, MqoFace& face, int cutType) const {
		Array<int> indices = face.indices;
		int vNum = indices.size();

		if (vNum == 3 || cutType == 0) {
			face.defaultCutting();
			return vNum;
		}

		// _̈ʒu
		Array<Vec3> positions = Array<Vec3>(vNum);
		for (int i = 0; i < vNum; ++i) {
			positions[i] = obj.vertices[indices[i]];
		}

		Array<Vec3> vs = Array<Vec3>(vNum);
		// K܂ł
		Array<Vec3> dirs = Array<Vec3>(vNum);
		Array<Vec3> faceDirs = Array<Vec3>(vNum);

		double maxScore = -2.0;
		// Jbg
		int funRoot = -1;
		for (int i = 0; i < vNum; ++i) {
			// xNǧvZ
			for (int j = 0; j < vNum - 1; ++j) {
				int j1 = (i + 1 + j) % vNum;
				vs[j] = positions[j1] - positions[i];
				dirs[j] = vs[j].normalized();
			}
			// ʖ@̌vZ
			for (int j = 0; j < vNum - 2; ++j) {
				faceDirs[j] = -dirs[j + 1].cross(dirs[j]);
				faceDirs[j].normalize();
			}
			// ς̌vZ
			double score = 2.0;
			for (int j = 0; j < vNum - 3; ++j) {
				double inp = faceDirs[j].dot(faceDirs[j + 1]);
				score = s3d::Min(score, inp);
			}

			if (score > maxScore) {
				maxScore = score;
				funRoot = i;
			}
		}

		// 򂷂
		if (vNum <= 5 || cutType == 1) {
			face.funCutting(funRoot);
			return vNum;
		}

		//// 2ˉeČ߂

		// xNǧ
		Vec3 bestDir = Vec3();
		int minCount = vNum;
		for (int i = 0; i < vNum - 2; ++i) {
			Vec3 sumDir = Vec3();
			int count = 0;
			for (int j = 0; j < vNum - 2; ++j) {
				double inp = faceDirs[i].dot(faceDirs[j]);
				sumDir += s3d::Math::Sign(inp) * faceDirs[j];
				if (inp == 0.0) {
					++count;
				}
			}

			if (count < minCount) {
				minCount = count;
				bestDir = sumDir;
				if (count == 0) {
					break;
				}
			}
		}

		bestDir.normalize();
		double x = bestDir.x;
		double y = bestDir.y;
		double z = bestDir.z;

		Vec3 xBase = { z, 0.0, -x };
		Vec3 yBase = { -x*y, x * x + z * z, -z*y };
		if (z == 0.0) {
			xBase = { 0.0, 0.0, -1.0 };
			yBase = { -y, x, 0.0 };
		}
		xBase.normalize();
		yBase.normalize();

		// ˉe
		Array<Vec3> projs = Array<Vec3>(vNum);
		for (int i = 0; i < vNum; ++i) {
			projs[i].x = positions[i].dot(xBase);
			projs[i].y = positions[i].dot(yBase);
			projs[i].z = positions[i].dot(bestDir);
		}


		// 2ŕ߂Ă
		Array<Vec3> angs = Array<Vec3>(vNum);
		// Ăʓ[J_Xg
		Array<int> subIndices = Array<int>(vNum);
		int n = vNum;
		for (int i = 0; i < n; ++i) {
			subIndices[i] = i;
		}

		face.tris.clear();

		for (int h = 0; h < vNum; ++h) {
			// ʏł̕xNg
			for (int i = 0; i < n; ++i) {
				int i1 = (i + 1) % n;
				vs[i] = projs[subIndices[i1]] - projs[subIndices[i]];
				vs[i].z = 0.0;
				dirs[i] = vs[i].normalized();
			}


			// pxƌ߂
			int posi = 0;
			int zero = 0;
			int nega = 0;
			double sumPosi = 0.0;
			double sumNega = 0.0;
			// #0 ̑Oɑ΂ x͓ρAy͊OςƂ

			for (int i = 0; i < n; ++i) {
				int ip = (i + n - 1) % n;
				double inp = -dirs[ip].dot(dirs[i]);
				double outp = -dirs[ip].cross(dirs[i]).z;
				angs[i].x = inp;
				angs[i].y = outp;
				angs[i].z = s3d::Math::Acos(inp);

				if (outp > 0.0) {
					++posi;
					sumPosi += angs[i].z;
				}
				else if (outp < 0.0) {
					++nega;
					sumNega += angs[i].z;
				}
				else {
					++zero;
				}
			}

			// ǂЕ̏ꍇ͓ʂȂ̂ŎRɕĂ悢
			if (posi == 0 || nega == 0) {
				// c낤
				for (int i = 0; i < n - 2; ++i) {
					Array<int> tri = { subIndices[0],
						subIndices[i + 1],
						subIndices[i + 2] };
					face.tris.push_back(tri);
				}
				break;
			}

			// sp̕
			double acute = 0.0;
			double angle = sumPosi + s3d::Math::TwoPi * nega - sumNega;
			angle = s3d::Math::Round(angle / s3d::Math::Pi);
			if (angle == (double)(n - 2)) {
				acute = 1.0;
			}
			else {
				angle = sumNega + s3d::Math::TwoPi * posi - sumPosi;
				angle = s3d::Math::Round(angle / s3d::Math::Pi);
				if (angle == (double)(n - 2)) {
					acute = -1.0;
				}
			}

			int target = -1;
			for (int i = 0; i < n; ++i) {
				if (acute != s3d::Math::Sign(angs[i].y)) {
					// 180xȏȂ炻̋ߕӂ̓AEg
					continue;
				}

				int i1 = (i + 1) % n;
				int ip = (i + n - 1) % n;

				bool flag = true;
				for (int j = 2; j < n - 1; ++j) {
					int i2 = (i + j) % n;

					Vec3 dir = projs[subIndices[i2]] - projs[subIndices[i]];
					double s = dir.dot(dirs[i]) / vs[i].length();
					double t = dir.dot(-dirs[ip]) / vs[ip].length();

					// Op`̓ɂAEg
					if (s >= 0.0 && t >= 0.0 && ((s + t) <= 1.0)) {
						// ł
						flag = false;
						break;
					}
				}

				if (flag) {
					target = i;
					break;
				}
			}

			if (target < 0) {
				target = 1;
			}

			// target Ō܂Ă̂2ɕ
			Array<int> tri = { subIndices[(target + n - 1) % n],
				subIndices[target],
				subIndices[(target + 1) % n] };
			face.tris.push_back(tri);
			// cčXV
			for (int i = target; i < n - 1; ++i) {
				subIndices[i] = subIndices[i + 1];
			}
			--n;
		}

		// ʖ@vZ
		for (int i = 0; i < vNum-2; ++i) {
			int i0 = face.tris[i][1];
			int i1 = face.tris[i][2];
			int ip = face.tris[i][0];

			Vec3 v0 = positions[i1] - positions[i0];
			Vec3 v1 = positions[i0] - positions[ip];
			faceDirs[i] = -v1.cross(v0);
			faceDirs[i].normalize();
		}

		double projScore = 2.0;
		// ʂ (vNum - 2) 
		for (int i = 0; i < vNum - 2; ++i) {
			for (int j = 0; j < vNum - 2; ++j) {
				if (i >= j) {
					continue;
				}

				int count = 0;
				for (int k0 = 0; k0 < 3; ++k0) {
					for (int k1 = 0; k1 < 3; ++k1) {
						count += (face.tris[i][k0] == face.tris[j][k1]);
					}
				}
				if (count >= 2) {
					double inp = faceDirs[i].dot(faceDirs[j]);
					if (inp < projScore) {
						projScore = inp;
						if (projScore < maxScore) {
							// ł؂
							break;
						}
					}
				}
			}
		}

		// rėǂgp
		if (projScore < maxScore) {
			face.funCutting(funRoot);
			return vNum;
		}

		// ˉeJbg̗p
		return vNum;
	}



	/// <summary>Object̏ԂvZglobal*֔fB
	/// scale͖ 1 1 1 Ƃ݂Ȃ</summary>
	bool computeBaseCoordinates() {
		int num = this->nodes.size();
		for (int i = 0; i < num; ++i) {
			ObjectNode& node = this->nodes[i];
			ObjectNode* p = &this->modelNode;
			int parent = node.parentIndex;
			if (parent >= 0 && parent < num) {
				p = &this->nodes[parent];
			}

			node.globalMov = p->globalMov
				+ p->globalRot * node.translation;
			node.globalRot = p->globalRot * node.rot;
		}
		return true;
	}


	/// <summary>̃CX^XێĂObjectNodeɑΉ\
	/// MotionNodezVɈB</summary>
	MotionInstance CreateMotionInstance() {
		MotionInstance ret;
		ret.modelNode = this->modelNode;
		int num = this->nodes.size();
		for (int i = 0; i < num; ++i) {
			const auto n = this->nodes[i];
			MotionNode node = n;
			ret.nodes.push_back(node);
		}
		return ret;
	}


	/// <summary>̃NXێĂPmdFile֕ϊ</summary>
	/// <param name="f">o͐</param>
	bool ConvertToPmd(PmdFile& f) {
		f.clear();

		Array<int> offsets;

		for (int i = 0; i < this->nodes.size(); ++i) {
			const ObjectNode& node = this->nodes[i];

			// ItZbg͂܂ł̒_
			offsets.push_back(f.vertices.size());

			// _
			int n = node.meshData.vertices.size();
			for (int j = 0; j < n; ++j) {
				const auto& mv = node.meshData.vertices[j];

				PmdVertex v;
				v.position = mv.position;
				v.normal = mv.normal;
				v.texcoord = mv.texcoord;
				v.bones[0] = i;
				v.bones[1] = i;
				v.weight = 100;
				v.edgeFlag = 0; // 0:ʏA1:GbW
				f.vertices.push_back(v);
			}

			// 1IuWFNg1{[
			PmdBone b;
			b.name = node.name;
			b.nameE = Format(L"bone", Pad(f.bones.size(), {5,L'0'}));
			b.parent = node.parentIndex;
			b.child = node.childIndex;
			b.position = node.globalMov;
			b.type = PMDBONE_ROTMOV;
			f.bones.push_back(b);
		}

		for (int i = 0; i < this->mats.size(); ++i) {
			for (int j = 0; j < this->nodes.size(); ++j) {
				if (i < this->nodes[j].groups.size()) {
					const auto& group = this->nodes[j].groups[i];
					for (int k = 0; k < group.indices.size(); ++k) {
						f.faces.push_back(group.indices[k] + offsets[j]);
					}
				}
			}
		}

		for (int i = 0; i < this->mats.size(); ++i) {
			Material2& mat = this->mats[i];

			PmdMaterial m;
			m.diffuse = mat.diffuse;
			m.power = mat.shininess;
			m.specular = mat.specular;
			m.ambient = mat.ambient;
			m.texName = mat.diffuseTextureName;
			if (mat.toonNumber >= 1) {
				m.toonIndex = mat.toonNumber - 1;
			}
			else {
				m.toonIndex = 255;
			}
			m.edgeFlag = mat.edge;

			m.matVertexNum = 0;
			for (int j = 0; j < this->nodes.size(); ++j) {
				const auto& node = this->nodes[j];
				if (i < node.groups.size()) {
					m.matVertexNum += node.groups[i].num();
				}
			}

			f.materials.push_back(m);
		}

		// wb_
		f.name = FileSystem::BaseName(this->path);
		f.nameE = L"Model name";
		f.comment = FileSystem::FileName(this->path) + L" ϊ";
		f.commentE = L"Comment.";

		// {[\g
		int adjustBoneNum = f.bones.size() - 1;
		if (adjustBoneNum < 0) {
			adjustBoneNum = 0;
		}
		int numInFrame = (adjustBoneNum <= 1200) ? 5 : 254;
		int frameNum = (adjustBoneNum + numInFrame - 1) / numInFrame;
		if (frameNum > 255) {
			frameNum = 255; // ő255
		}
		int index = 1; // 擪܂܂Ȃ
		for (int i = 0; i < frameNum; ++i) {
			PmdFrame frame;
			String s = Format(L"bones", Pad(i+1, { 3, L'0' }));
			frame.name = s;
			frame.nameE = s;
			for (int j = 0; j < numInFrame; ++j) {
				if (index >= f.bones.size()) {
					break;
				}
				frame.bones.push_back(index);
				++index;
			}
			f.frames.push_back(frame);
		}

		return true;
	}
};
